//------------------------------------------------------------------------
//
// OpenDCC - OpenDecoder2 - SignalDecoder2
//
// Copyright (c) 2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      config.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-02-14 V0.1 kw start
//            2007-04-01 V0.2 kw added variables for Servo Curve
//            2007-05-07 V0.3 kw added Servo Repeat
//                               and parameters for multi positioning
//            2007-06-18 V0.4 kw added C_Tick
//            2007-09-18 V0.5 kw added extra CV for DMX (gap),
//                               no functional change
//            2007-09-26 V0.6 kw moved to new project (Signaldecoder2)
//
//------------------------------------------------------------------------
//
// purpose:   flexible general purpose decoder for dcc
//            This is the central definition file for the project
//
//------------------------------------------------------------------------
//
// content:   1. Project settings
//               1.a) General
//               1.b) Configuration of Software Modules
//
//            2. EEPROM definitions (CV's)
//            3. Global variables
//            4. Useful inline routines (delay, Semaphor-operations)
//
//========================================================================
//
#ifndef _CONFIG_H_
#define _CONFIG_H_

//========================================================================
// 1. Project Definitions
//========================================================================
//
// 1.a) General
//
#define OPENDECODER_VERSION   0x20

// Definition of target hardware:
//   
#define OPENDECODER2 2
//  OPENDECODER2:  this board has:
//                 - ATmega8515 @ 8MHz
//                 - 2 Servo connectors
//                 - DMX (Tx and Rx),
//                 - 8 Powerports with integrated feedback
//                 see: www.opendcc.de
//
#define OPENDECODER3 3
//  OPENDECODER3:  this board has:
//                 - ATmega162 @ 8MHz (2 Uarts)
//                 - 2 Servo connectors with polyfuse
//                 - complete BiDi / railcom interface
//                 - optional: DMX, Xpressnet, Loconet
//                 - 4 relais out, 4 tracers inputs
//                 see: www.opendcc.de and www.tams-online.de

#define TARGET_HARDWARE		OPENDECODER2

#define FALSE  0
#define TRUE   (!FALSE)

// Debugswitches:  0: no debugging, interrupts running -> alive!
//                 1: no ints, some lowlevel test

#define DEBUG  0

// 1.b) Configuration of Software Modules
//
// Note: depending on the desired target decoder, you may disable one of the
//       following enables.
//       Due to EEPROM and FLASH limitations, either SERVO or DMX *must* be disabled.
// 

//#define PORT_ENABLED	FALSE			// FALSE: PortEngine only does the timers
#define PORT_ENABLED	TRUE		// TRUE: PortEngine does control the outputs

#define SIMULATION  0               // 0: real application
                                    // 1: test receive routine
                                    // 2: test timing engine
                                    // 3: test action

//========================================================================
// 2. EEPROM Definitions (CV's)
//========================================================================
//
// content is defined in config.c

#define CV_REMAPPING   TRUE              // if true: remap cv 513 to cv 1 aso;

#include "cv_define.h"

extern t_cv_record CV EEMEM;

extern t_cv_record CV_PRESET PROGMEM;

//========================================================================
// 3. Global variables
//========================================================================

//---------------------------------------------------------------------
// Timing Definitions:
//

#define TICK_PERIOD 20000L       // 20ms tick for Timing Engine
                                 // => possible values for timings up to
                                 //    5.1s (=255/0.020)
                                 // note: this is also used as frame for
                                 // Servo-Outputs (OC1A and OC1B)

//----------------------------------------------------------------------------

extern volatile signed char timerval;    // gets incremented in the timetick



//========================================================================
// 4. Project Definitions
//========================================================================
//

//-------------------------------------------------------------------------
// 4.a) Definitions for inter process communication
//-------------------------------------------------------------------------
// These routines are *inline*, so we keep them in the header.
//
// usage:   semaphor_set(flag, state) to do the communication.
//          semaphor_get(flag)

extern volatile unsigned char Communicate;

#define C_Received		0		// a new DCC was received - issued by ISR(Timer1)
													//                          cleared by main
#define C_DoSave			1		// a new PORT state should be saved
													//                        - issued by action
													//                          cleared by main
#define C_Tick				2		// a Tickevent happened

//========================================================================
// 4.b) Useful inline code
//========================================================================

static inline unsigned char semaphor_query(unsigned char flag) __attribute__((always_inline));
unsigned char semaphor_query(unsigned char flag)
{
	return (Communicate & (1<<flag));
}
               
static inline void semaphor_set(unsigned char flag) __attribute__((always_inline));
void semaphor_set(unsigned char flag)
{
	cli();
	Communicate |= (1<<flag);
	sei();
}
       
static inline unsigned char semaphor_get(unsigned char flag) __attribute__((always_inline));
unsigned char semaphor_get(unsigned char flag)
{
	unsigned char value;
	cli();
	value = Communicate & (1<<flag);
	Communicate &= (~(1<<flag));
	sei();
	return(value);
}

//------------------------------------------------------------------------
// Delay-Macro (all values in us) -> this is busy waiting
//------------------------------------------------------------------------
//
// same macro as in util/delay.h, but use up to some ms.
// The maximal possible delay is 262.14 ms / F_CPU in MHz.
// This is 16ms for 16MHz; longest used delay: 1000us

#ifndef _UTIL_DELAY_H_
  #include <util/delay.h>
#endif

static inline void _mydelay_us(double __us) __attribute__((always_inline));
void _mydelay_us(double __us)
{
	uint16_t __ticks;
	double	 __tmp = ((F_CPU) / 4e6) * __us;

	if (__tmp < 1.0)
		__ticks = 1;
	else if (__tmp > 65535)
		__ticks = 0;    /* i.e. 65536 */
	else
		__ticks = (uint16_t)__tmp;
	_delay_loop_2(__ticks);
}   

static inline void _restart(void) __attribute__((always_inline));
void _restart(void)
{
	cli();

	// laut diversen Internetseiten sollte folgender Code laufen -
	// tuts aber nicht, wenn man das Assemblerlistung ansieht.
	// void (*funcptr)( void ) = 0x0000;    // Set up function pointer
	// funcptr();                        // Jump to Reset vector 0x0000

	__asm__ __volatile 
	(
		"ldi r30,0"	"\n\t"
		"ldi r31,0"	"\n\t"
		"icall"			"\n\t"
	);
}

#endif	// _CONFIG_H_
