//------------------------------------------------------------------------
//
// OpenDCC - OpenDecoder2
//
// Copyright (c) 2006,2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      dimm_engine.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-02-20 V0.1  kw copied from opendecoder.c
//                                removed code for versatile output -
//                                we only do turnout commands and permanent
//                                commands (up to now)
//            2007-04-10 V0.2  kw added init code for timer 1
//                                turn off output only if time > 0
//                                outputs with time = 0 are kept forever
//            2007-04-23 V0.3  kw added compile switch
//                                PORT_ENABLED
//            2007-04-27 V0.4  kw evaluation of ReceivedActivate
//            2007-05-14 V0.5  kw added masking of TOIE1 to prevent reentry
//            2007-05-25 V0.6  kw Bugfix in init of FEEDBACK_PULLUP
//            2007-10-29 V0.7  kw added belgian Signals
//            2007-05-30 V0.8  kw added reinit of states (except blinking)
//            2008-11-23 V0.9  kw CurrentBlink added
//            2009-01-13 V0.10 kw bugfix in light_column
//
// tests:     2007-04-14 kw: feedback tested, FBM = 0,1; Magnet coils
//
//------------------------------------------------------------------------
//
// purpose:   flexible general purpose decoder for dcc
//            here: timing engine for outputs
//                  software pwm and state engine
//
// content:   A DCC-Decoder for ATmega8 and other AVR
//
//             1. Defines and variable definitions
//             2. Init
//             3. ISR
//             4. Predefined values
//             5. action - performing the Command from the DCC call
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "myeeprom.h"            // just a wrapper function
#include "config.h"
#include "hardware.h"
#include "dcc_receiver.h"
#include "main.h"
#include "dimm_engine.h"

#define DIMM_ENABLED  TRUE

//=============================================================================
// 1. Definitions
//=============================================================================
//
// Output:
// Port Engine does the control of the outputs only if PORT_ENABLED is true
//
#ifndef PORT_ENABLED
  #warning PORT_ENABLED not defined, now enabled by default (should be done in config.h)
  #define PORT_ENABLED  TRUE
#endif

// Definitions for the DIMM-Engine
// see below on information about the DIMM-Engine
//
#define PWM_TICK_PERIOD			100L			// 100us tick for PWM Engine
#define PWM_STEPS						50L
#define PWM_PERIOD					(PWM_TICK_PERIOD * PWM_STEPS)		// 5.0ms = 100 * 50L
#define DIMM_DOWN_SAME_BULB	1					// 1: wenn im neuen Fahrbegriff die Lampe 
																			//    wieder leuchtet, zuerst abdimmen,
																			//    dann wieder aufdimmen.
																			// 0: leuchtet durchgehend  

#define DIMM_RANGE_MIN	100				    // aktiver Bereich 100-150
#define DIMM_RANGE_MAX	(DIMM_RANGE_MIN + PWM_STEPS + 1)

#define PWM_TABLE_SIZE	(PWM_STEPS)		// Grsse der PWM-Tabelle

// Excel: =POTENZ(1.083110727; Ax) + 100		(x = 0..49)
uint8_t pwmTable[PWM_TABLE_SIZE] PROGMEM = {	// 50 exponentielle Werte
	100, 101, 101, 101, 101, 101, 102, 102, 102, 102, 102, 102, 103, 103, 103, 103, 104, 
	104, 104, 105, 105, 105, 106, 106, 107, 107, 108, 109, 109, 110, 111, 112, 113, 114, 
	115, 116, 118, 119, 121, 123, 124, 126, 129, 131, 134, 136, 139, 143, 146, 150 };

#define LED_PAUSE			2100000L / TICK_PERIOD;
#define LED_OFFTIME		 720000L / TICK_PERIOD;
#define LED_ONTIME		 360000L / TICK_PERIOD;
#define LED_REST			 360000L / TICK_PERIOD;

// local Declaration
void set_new_light_val_all(unsigned char signNbr, unsigned char valid, unsigned char pattern, unsigned char blink);
void SPI_Output(uint8_t signNbr);

//----------------------------------------------------------------------------
// Global Data
// out_pwm controls outputs, monoflop pulses and toggling operations
//------------------------------------------- PWM

// a) running values
unsigned char cur_dimm;
unsigned char CurrentTarget[NUM_SIGNAL];
unsigned char CurrentBlink[NUM_SIGNAL];

// b) target values
volatile struct
{
	unsigned char rest;					// Zeit bis zum nchsten Wechsel in PWM_PERIOD (5ms)
	unsigned char ontime;   		// Einschaltzeit
	unsigned char offtime;      // Ausschaltzeit
	unsigned char pwmTabInx;		// aktueller Istwert (Index) der PWM-Tabelle
	unsigned char light_A_val;  // aktueller Zielwert in Richtung A
	unsigned char delta_A;      // aktuelles Increment in Richtung A
	unsigned char light_B_val;  // aktueller Zielwert in Richtung B
	unsigned char delta_B;      // aktuelles Decrement in Richtung B
	unsigned char curPattern;		// der aktuelle Fahrbegriff
	unsigned char newPattern;		// der neue Fahrbegriff
	unsigned char displayed;		// die angezeigten Signal-Daten (Fahrbegriff)
} out_pwm[NUM_SIGNAL];

unsigned char	DimmUpDelay;		// given in Ticks (and loaded from eeprom);
unsigned char	BlinkPeriod_2;	// if a blink bit is set, use this value to determine period

// some static used variables
unsigned char	Display_MainSign;		// HauptSignal anzeigen
unsigned char	Display_PrevSign;		// VorSignal anzeigen

//----------------------------
volatile struct
{
	unsigned char rest;				// Zeit bis zum wechsel in Ticks (20ms)
	unsigned char ontime;			// Einschaltzeit
	unsigned char offtime;		// Ausschaltzeit
	unsigned char pause;			// Pausenzeit
	unsigned char flashes;		// Anzahl Pulse
	unsigned char act_flash;    
} led;

void turn_led_on(void)
{
	led.rest = 0;
	LED_ON;
}

void turn_led_off(void)
{
	led.rest = 0;
	LED_OFF;
}

// make a series of flashes, then a longer pause
void flash_led_fast(unsigned char count)
{
	led.act_flash = 1;
	led.flashes		= count;
	led.pause			= LED_PAUSE;
	led.offtime		= LED_OFFTIME;
	led.ontime		= LED_ONTIME;
	led.rest			= LED_REST;
	LED_ON;
}

// Trick:   Speed up port access by declaring port access as inline code
//          This is done with a subroutine; if gcc runs with -os, this results in
//          single cbi and sbi statements!
// Note:    OUTPUT_PORT is defined in hardware.h
//
static inline void output(uint8_t* port, uint8_t pin_no, uint8_t state) __attribute__((always_inline));
void output(uint8_t* port, uint8_t pin_no, uint8_t state)
{
	if (state == 0) {
		*port &= (~(1<<pin_no));
	} else {
		*port |= (1<<pin_no);
	}
}

static inline void disable_timer_interrupt(void) __attribute__((always_inline));
void disable_timer_interrupt(void)
{
	TIMSK &= ~(1<<OCIE1A);	// Timer1 Compare A
}

static inline void enable_timer_interrupt(void) __attribute__((always_inline));
void enable_timer_interrupt(void) 
{
	TIMSK |= (1<<OCIE1A);		// Timer1 Compare A
}

//==============================================================================
// Section 2
//------------------------------------------------------------------------------
//
// init_dimm_engine(void)
//   initializes the output of the decoder 
//   a) setup timer (with TICK_PERIOD)
//   b) load values for port states from eeprom
//   c) preload outputs

void init_dimm_engine(void)
{
	unsigned char j, sign;

	#if (DIMM_ENABLED == TRUE)
    // Init Timer1 as CTC 
    // check PWM_TICK_PERIOD and F_CPU
		#define TIMER1_CLKDIV 8              // possible values: 1, 8, 64, 256, 1024

		#if (F_CPU / 1000000L * PWM_TICK_PERIOD / TIMER1_CLKDIV) > 255L
			#warning: overflow in OCR1A - check TICK_PERIOD and F_CPU
			#warning: suggestion: use a larger clkdiv
		#endif
		#if (F_CPU / 1000000L * PWM_TICK_PERIOD / TIMER1_CLKDIV) < 30L
			#warning: resolution accuracy in OCR1A too low - check TICK_PERIOD and F_CPU
			#warning: suggestion: use a smaller clkdiv
		#endif

		OCR1A	 = F_CPU / 1000000L * PWM_TICK_PERIOD / TIMER1_CLKDIV;	// 200

		TCCR1A = (0 << COM1A1)		// compare match A (disconnected)
           | (0 << COM1A0) 
           | (0 << COM1B1)		// compare match B (disconnected)
           | (0 << COM1B0)
           | (0 << FOC1A)			// reserved
           | (0 << FOC1B)			// reserved
           | (0 << WGM11)
           | (0 << WGM10);		// Timer0 Mode 4 = CTC - Int on compare A (Top = OCR1A)

		TCCR1B = (0 << ICNC1)
           | (0 << ICES1)
           |	0								// reserved
           | (0 << WGM13)
           | (1 << WGM12)

    #if   TIMER1_CLKDIV == 1
           | (0 << CS12) | (0 << CS11)  | (1 << CS10);		// CS1[2:0]=001 clkdiv 1
    #elif TIMER1_CLKDIV == 8
           | (0 << CS12) | (1 << CS11)  | (0 << CS10);		// CS1[2:0]=010 clkdiv 8
    #elif TIMER1_CLKDIV == 64
           | (0 << CS12) | (1 << CS11)  | (1 << CS10);		// CS1[2:0]=011 clkdiv 64 
    #elif TIMER1_CLKDIV == 256
           | (1 << CS12) | (0 << CS11)  | (0 << CS10);		// CS1[2:0]=100 clkdiv 256
    #elif TIMER1_CLKDIV == 1024
           | (1 << CS12) | (0 << CS11)  | (1 << CS10);		// CS1[2:0]=101 clkdiv 1024   
    #else
			#warning: TIMER1_CLKDIV is void
    #endif

		TIMSK |= 0
					|	 0
					|	(0<<TICIE1)				// Timer1 Input Capture
					|	(1<<OCIE1A)				// Timer1 Compare A
					|	(0<<OCIE1B)				// Timer1 Compare B
					|	(0<<TOIE1)				// Timer1 Overflow (-> was 1)
					|	 0
					|	 0;

		// now init all state variables
		timerval = 0;
		cur_dimm = DIMM_RANGE_MIN;

		// Initialize SPI
		#define SPI_CLKDIV 4		// possible values: 4, 16, 64, 128
		// F_CPU / (SPI_CLKDIV / SPI2X) = 0.125us
		SPCR	= (0 << SPIE)			// SPI Interrupt Enable
					| (1 << SPE)			// SPI Enable
					| (0 << DORD)			// Data Order, MSB of the data word is transmitted first
					| (1 << MSTR)			// Master
					| (0 << CPOL)			// Clock Polarity, SCK is low when idle
					| (0 << CPHA)			// Clock Phase, Trailing Edge setup

		// SPI Frequency
		#if   SPI_CLKDIV == 4
					| (0 << SPR1) | (0 << SPR0);		// SPR[1:0]=00 clkdiv 4
		#elif SPI_CLKDIV == 16
					| (0 << SPR1) | (1 << SPR1);		// SPR[1:0]=01 clkdiv 16
		#elif SPI_CLKDIV == 64
					| (1 << SPR1) | (0 << SPR1);		// SPR[1:0]=10 clkdiv 64 
		#elif SPI_CLKDIV == 128
					| (1 << SPR1) | (1 << SPR1);		// SPR[1:0]=11 clkdiv 128
		#else
			#warning: SPI_CLKDIV is void
		#endif

		SPSR	= (1 << SPI2X);			// Double SPI Speed Bit

	#endif	// DIMM_ENABLED

	// b) load values for port states from eeprom

	// Delta fr Glhlampensimulation vorbelegen

	j = my_eeprom_read_byte(&CV.DimmUpSpeed);
	if (j == 0)
		j = 1;
	if (j > PWM_STEPS + 2)
		j = PWM_STEPS + 2;

	for (sign = 0; sign < NUM_SIGNAL; sign++) {
		out_pwm[sign].delta_A = j;
	}

	j = my_eeprom_read_byte(&CV.DimmDownSpeed);
	if (j == 0)
		j = 1;
	if (j > PWM_STEPS + 2)
		j = PWM_STEPS + 2;

	for (sign = 0; sign < NUM_SIGNAL; sign++) {
		out_pwm[sign].delta_B = j;
	}

	j = my_eeprom_read_byte(&CV.TargetUp);				// add Offset for engine
	if (j >= PWM_TABLE_SIZE)
		j = PWM_TABLE_SIZE - 1;
	for (sign = 0; sign < NUM_SIGNAL; sign++) {
		out_pwm[sign].light_A_val = j;
	}

	j = my_eeprom_read_byte(&CV.TargetDown);			// add Offset for engine
	for (sign = 0; sign < NUM_SIGNAL; sign++) {
		out_pwm[sign].light_B_val = j;
		out_pwm[sign].pwmTabInx = 0;
	}

	DimmUpDelay		= my_eeprom_read_byte(&CV.DimmUpDelay);
	BlinkPeriod_2 = my_eeprom_read_byte(&CV.BlinkPeriod_2);

	// c) preload outputs and states

	for (sign = 0; sign < NUM_SIGNAL; sign++) {
		CurrentTarget[sign] = my_eeprom_read_byte(&CV.LastPortState[sign]);
		CurrentBlink[sign]	= my_eeprom_read_byte(&CV.LastBlinkState[sign]);
	}
	Display_MainSign = my_eeprom_read_byte(&CV.DsplyMainSign);	// HauptSignal anzeigen
	Display_PrevSign = my_eeprom_read_byte(&CV.DsplyPrevSign);	// VorSignal anzeigen

	switch (my_eeprom_read_byte(&CV.LightMode)) {
		default:
		case  2: // 02 = 16 aspects, dimmed, static states
			for (sign = 0; sign < NUM_SIGNAL; sign++) {
				set_new_light_val_all(sign, 0xFF, CurrentTarget[sign], CurrentBlink[sign]);
			}
			break;
	} 
}	// init_dimm_engine

//==============================================================================
//
// Section 3
//
// Timing Engine
//
// Howto:    Timer1 (with prescaler 8 and 16 bit total count) triggers
//           an interrupt every PWM_PERIOD (=100us @16MHz);
// 
//==============================================================================
//
// Section 3
//
// Lichtsteuerung als PWM-Dimmer
//
//             |                        |                          |
// Output:     |XXXXXXXXXX______________|XXXXXXXXXX________________|
//             |<-------->              |<------------------------>|
//             | pwmTabInx              |        PWM_PERIOD        |
//         ----|------------------------|--------------------------|---> Time
//
//
// 1. Einstellen des aktuellen Helligkeitswertes (Dimmer)
//
//    Es gibt 50 Helligkeitstufen.
//    Alle 100us erfolgt ein Interrupt, dieser schaltet den Dimmer um eine
//    Stufe weiter, nach 50 Stufen wird wieder von vorne begonnen.
//    Stufe MIN:  alle Ports mit einem pwmTabInx > DIMM_MIN werden eingeschaltet.
//    Stufe x:    Ein Port, dessen pwmTabInx kleiner x ist, wird abgeschaltet.
//    Stufe MAX:  Restart und Meldung an den DIMMER (C_Tick)
//
//    Folge: Alle Ports mit einem pwmTabInx kleiner DIMM_RANGE_MIN sind
//    dauerhaft aus, alle Ports mit einem pwmTabInx grsser DIMM_RANGE_MAX
//    sind dauerhaft ein.
//
//    Da der PWM 50 Stufen hat und alle 100us ein Interrupt erfolgt, wird dieser
//    Durchlauf alle 5.0ms durchgefhrt. Diesen teilen sich die vier Signalpaare
//		zu je 20ms. Pro Signalpaar entspricht dies einer Refreshrate von 50Hz.
//
// 2. Langsame Vernderung der Helligkeit
//
//    Nach einem Zyklus des PWM wird vom Hauptprogramm der neue aktuelle
//    pwmTabInx ausgerechnet. Hierzu wird vom aktuellen Wert mit einem
//    Schritt "delta" nach oben oder unten gerechnet, bis der neue Zielwert
//    erreicht ist.
//
//    Die Zykluszeit der PWM pro Signalpaar ist 20ms, somit wird je nach
//		"delta" folgende Dimmzeit erreicht: (PWM-Stufen * Zykluszeit / delta)
//
//      delta    |    Dimmzeit
//    -----------------------------
//        1      |    1000ms
//        2      |     500ms
//        3      |     333ms
//        4      |     250ms
//        5      |     200ms
//        6      |     167ms
//      100      |    sofort
//
//    Der neue Zielwert wird in light_val hinterlegt. 
//
//    Wenn man von einem Wert kleiner DIMM_RANGE_MIN startet, dann wird
//    der Port erst mit Verzgerung aufgedimmt, weil zuerst der Bereich
//    bis DIMM_RANGE_MIN "aufgedimmt" wird.
//    Dies wird dazu benutzt, zuerst das alte Signalbild wegzudimmen 
//    und dann das neue Signalbild aufzudimmen.
//
// 3. Signalbilder
// 
//    Signalbilder werden als Bitfeld hinterlegt. Fr jedes Kommando (Aspect)
//    gibt es ein Bitfeld, in dem das neue Signalbild abgelegt ist und eine
//    Gltigkeitsmaske, diese bestimmt, auf welche Dimmwerte das Signalbild
//    wirken soll. Mit diesen beiden Pattern wird "set_new_light_val_all" 
//    aufgerufen.
//
// 4. Blinken
//    
//    Falls ontime bzw. offtime ungleich 0 sind, wird nach Ablauf der
//    jeweils andere Phasenwert geladen.
//

// This is the PWM Interrupt

ISR(TIMER1_COMPA_vect)                    // Timer1 Compare Int, macht PWM
//====================
// Wird alle 100us aufgerufen
{
  static uint8_t signNbr = 0;
	uint8_t pwmTabVal;

  disable_timer_interrupt();
  sei();                                  // allow dcc receive
  cur_dimm++;

	pwmTabVal = pgm_read_byte(&pwmTable[out_pwm[signNbr].pwmTabInx]);

  if (cur_dimm == DIMM_RANGE_MAX) {				// DIMM_RANGE_MAX == 151
    cur_dimm = DIMM_RANGE_MIN;
		if (pwmTabVal > DIMM_RANGE_MIN) {
			out_pwm[signNbr].displayed = out_pwm[signNbr].curPattern;	// Einschalten
		}
    Communicate |= (1<<C_Tick);
    timerval++;														// advance global tick
  } else {
		if (cur_dimm > pwmTabVal) {
			out_pwm[signNbr].displayed = 0;			// Ausschalten
		}
	}

	SPI_Output(signNbr);
	signNbr = (signNbr < NUM_SIGNAL - 1) ? signNbr + 1: 0;
	enable_timer_interrupt();
}

void SPI_Output(uint8_t signNbr)
//=============
{
	// SPI routine, Durchlaufzeit < 20us
	// Erstes Byte sind die Zeilen der Matrix	(Fahrbegriff)
	// Zweites Byte sind die Spalten der Matrix	(Signalnummer)
	uint8_t data;
	SHIFT_OE_OFF;												// Register-Ausgnge abschalten ~OE
	SHIFT_SCLR_ON;											// Reset-Impuls Schieberegister ~SCLR
	SHIFT_SCLR_OFF;

	// ertes Byte senden (Fahrbegriff)
	SPDR = out_pwm[signNbr].displayed;

	// zweites Byte senden (Matrix-Spalten)
	// Daten schon mal bereitstellen, whrend erstes byte noch ausgegeben wird.
	data = 0;
	if (Display_MainSign & (1<<signNbr))
		data |= (1<<signNbr);							// HauptSignal anzeigen
	if ((out_pwm[signNbr].displayed != 0xC0) || (Display_PrevSign & (1<<signNbr)))
		data |= ((1<<signNbr) << 4);			// VorSignal anzeigen
  while(!(SPSR & (1<<SPIF)));					// warten auf fertig (1. Byte)
	SPDR = data;
  while(!(SPSR & (1<<SPIF)));					// warten auf fertig (2. Byte)

	// Die Daten von den Schieberegistern in die Ausgaberegister bernehmen.
	// Dazu am RCK Eingang der Schieberegister einen 0-1-0 Puls erzeugen.
	SHIFT_RCK_ON;
	SHIFT_RCK_OFF;
	SHIFT_OE_ON;												// Register-Ausgnge wieder einschalten
} // SPI_Output

//==============================================================================
//
// Section 5
//
// Performing the Command

//------------------------------------------------------------------------------
// This Routine is called when myAdr is received

void set_new_light_val_all(unsigned char sign, unsigned char valid,
													 unsigned char pattern, unsigned char blink)
// Diese Funktion setzt die neuen Ziel-Bits fr DIMM
// beim Aufdimmen gibt es immer 400ms Verzgerung, damit ein Abdimmen vorher fertig wird.
{
	if (pattern == out_pwm[sign].newPattern)							// Doppelten Fahrbefriff unterdrcken
		return;
	out_pwm[sign].curPattern = out_pwm[sign].newPattern;	// Letzten Fahrbefriff sichern

	if (valid) {
		out_pwm[sign].newPattern = pattern;
		if (blink) {
			CurrentBlink[sign]		= blink;
			out_pwm[sign].rest		= BlinkPeriod_2;
			out_pwm[sign].ontime	= BlinkPeriod_2;
			out_pwm[sign].offtime	= BlinkPeriod_2;

			if (pattern) {
				CurrentTarget[sign]	= 0;				// Target is off, this will turn on after .rest
				out_pwm[sign].curPattern = pattern;
			} else {
				CurrentTarget[sign]	= pattern;	// Target is on, this will turn off after .rest
			}
		} else {		// static pattern
			CurrentBlink[sign] = 0;
			if (pattern) {
				// turn on
				#if DIMM_DOWN_SAME_BULB == 1
					CurrentTarget[sign]		= 0;
					out_pwm[sign].rest		= DimmUpDelay;		// original: DIMM_UP_DELAY / PWM_PERIOD;   
					out_pwm[sign].ontime	= 0;
					out_pwm[sign].offtime	= 0;
				#else
					if (CurrentTarget[sign]) {
						// already on, nothing to do, keep burning 
					} else {
						CurrentTarget[sign]		= 0;
						out_pwm[sign].rest		= DimmUpDelay;	// original: DIMM_UP_DELAY / PWM_PERIOD;   
						out_pwm[sign].ontime	= 0;
						out_pwm[sign].offtime	= 0;
					}
				#endif
			} else {	// turn off
				CurrentTarget[sign]		= 0;
				out_pwm[sign].rest		= 0;   
				out_pwm[sign].ontime	= 0;
				out_pwm[sign].offtime	= 0;
			}
		}
	} else {
		out_pwm[sign].newPattern = 0;
	}
}

//---------------------------------------------------------------------------------
// Dimmer()
// this routine is called from main(), if C_Tick is activated  
//
// Timing Engine
//
// Howto:    
// 1. Generelles Timing:
//    Diese Routine wird alle PWM_PERIOD aufgerufen. Es wird folgendes geprft:
//    a) Wenn out_pwm[port].rest gleich 0: dann bleibt dieser Port unverndert.
//    b) out_pwm[port].rest wird decrementiert, wenn es dabei 0 wird, dann
//       wird ein Dimmvorgang in die andere Richtung eingeleitet.
//
// 2. Dimm-bergnge:
//    Je nach aktueller Richtung des Dimmvorgang (CurrentTarget) wird der aktuelle
//    Dimmwert erhht oder erniedrigt.
//    Die Dimmrampe ist unabhngig von den Zeiten, die bei ontime bzw. offtime
//    vorgegeben werden.
//    Wenn ein Ausgang ohne PWM durchschalten soll, dann muss sein Delta sehr
//    gross gewhlt werden! (>50)

void Dimmer(void)
{
	unsigned char sign, my_rest, pwmTabInx;

	for (sign = 0; sign < NUM_SIGNAL; sign++) {
		my_rest = out_pwm[sign].rest;			// use a local variable to force
		if (my_rest != 0) {								// compiler to tiny code
			if (--my_rest == 0) {
				if (CurrentTarget[sign]) {		// bit was on
					my_rest = out_pwm[sign].offtime;
					CurrentTarget[sign] = 0;
				} else {
					my_rest = out_pwm[sign].ontime;
					CurrentTarget[sign] = out_pwm[sign].newPattern;
				}
			}
			out_pwm[sign].rest = my_rest;
		} else {													// 'rest' abgearbeitet, neuen Fahrbegriff bernehmen
			out_pwm[sign].curPattern = out_pwm[sign].newPattern;
		}

		pwmTabInx = out_pwm[sign].pwmTabInx;
		if (CurrentTarget[sign]) {				// we are in phase A -> incr
			if ((pwmTabInx + out_pwm[sign].delta_A) <= out_pwm[sign].light_A_val)
				pwmTabInx += out_pwm[sign].delta_A;
			else
				pwmTabInx = out_pwm[sign].light_A_val;
		} else {													// we are in phase B -> decr
			if ((int8_t)(pwmTabInx - out_pwm[sign].delta_B) >= out_pwm[sign].light_B_val)
				pwmTabInx -= out_pwm[sign].delta_B;
			else
				pwmTabInx = out_pwm[sign].light_B_val;
		}
		out_pwm[sign].pwmTabInx = pwmTabInx;
	}

	// now process led

	if (led.rest != 0) {
		if (--led.rest == 0) {
			if (LED_STATE) {
				if (led.act_flash == led.flashes) {
					led.rest = led.pause;
					led.act_flash = 0;
				} else {
					led.rest = led.offtime;
				}
				LED_OFF;
			} else {
				led.act_flash++;
				led.rest = led.ontime;
				LED_ON;
			}
		}
	}
}

//==============================================================================
//
// Section 4
//
// Pattern for different signals (predefined in PROGMEM, read with pgm_read_byte)
//

//==============================================================================
//
// Section 5
//
void set_signal_aspects(unsigned char aspect, unsigned char signNbr)
{
	switch (my_eeprom_read_byte(&CV.SignalType)) {
		default:
		case 0:		// 0 = User Defined
			if (aspect < my_eeprom_read_byte(&CV.NoOfAspects))
				set_new_light_val_all(signNbr, my_eeprom_read_byte(&CV.UDEF[aspect].Valid),
																			 my_eeprom_read_byte(&CV.UDEF[aspect].Pattern),
																			 my_eeprom_read_byte(&CV.UDEF[aspect].Blink));
			break;
	}
}

//------------------------------------------------------------------------------
// This Routine is called when myAdr is received

void action(unsigned int aspect, unsigned char signNbr)
//=========
{
	unsigned char lightMode;

	if (aspect > 15 || signNbr > 3)		// aspect or signNbr out of range
		return;													// not allowed
	lightMode = my_eeprom_read_byte(&CV.LightMode);

	if (lightMode > 0) {
		switch (lightMode) {
			default:
			case  2:	// 02 = 16 addresses, dimmed, static states
				set_signal_aspects(aspect, signNbr);
				break;
		}   
		semaphor_set(C_DoSave);
	}
}

void simple_preset(unsigned char command)
{
	// during programming with tracer, we do a mode select depending on the received
	// command.

	// two different setups:
	// with Jumper: Signal States
	// without Jumper: Turnout Decoder

	if (JUMPER_FITTED) {
		my_eeprom_write_byte(&CV.LightMode,	 2);	// 16 aspects, dimmed, static states
		my_eeprom_write_byte(&CV.SignalType, 0);
	} else {
		my_eeprom_write_byte(&CV.LightMode,	 0);	// standard
		my_eeprom_write_byte(&CV.SignalType, 0);
	}
	do {} while (!eeprom_is_ready());					// wait for write to complete
}


#if (PRBS_CODE == 1)
// linear feedback shift register (prbs)
//
//      |---|    |---|    |---|    |---|    |---|    |---|    |---|    |---|
//    ->| 0 |--->| 1 |-o->| 2 |-o->| 3 |-o->| 4 |--->| 5 |--->| 6 |--->| 7 |--o--->
//   |  |---|    |---| |  |---| |  |---| |  |---|    |---|    |---|    |---|  |
//   |                 |        |        |                                    |
//    <--------------- + <----- + <----- + <----------------------------------
//
unsigned char prbs8(unsigned char seed)
{
	unsigned char new_rnd;

	new_rnd = seed;                // copy bit 1
	new_rnd = new_rnd << 1;
	new_rnd = new_rnd ^ seed;      // xor bit 2
	new_rnd = new_rnd << 1;
	new_rnd = new_rnd ^ seed;      // xor bit 3
	new_rnd = new_rnd << 4;
	new_rnd = new_rnd ^ seed;      // xor bit 7

	// now put this bit to seed's lsb
	new_rnd = new_rnd >> 7;
	seed = seed << 1;
	new_rnd = new_rnd + seed; 
	return(new_rnd);
}
#endif
