//------------------------------------------------------------------------
//
// OpenDCC - OpenDecoder2
//
// Copyright (c) 2006, 2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      main.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-02-14 V0.1 kw copied from opendecoder.c
//                               split up the code in modules
//                               added routines for CV-Programming
//            2007-02-25 V0.2 kw corrected some items with Decoding
//            2007-04-10 V0.3 kw moved Timerinit to port_engine.c
//            2007-04-27 V0.4 kw all-in-one software
//            2007-05-31 V0.5 kw added Hardware 3, added Jumperquery
//                               for programming
//            2007-08-05 V0.6 kw change to CV struct 
//            2007-08-18 V0.7 kw added 5 flashes  if unprogrammed
//            2008-05-26 V0.8 kw added save of CurrentTarget if SignalDecoder
//            2008-05-26 V0.9 kw added CheckSave - saves CurrentTarget after 3s
//                               and hauptsignal_state + vorsignal_state
//            2008-11-23 V0.3 kw BlinkState neu dazu
//                               eeprom wrapper neu dazu!
//
// 2do: manual adjustment, tracers
//
//------------------------------------------------------------------------
//
// purpose:   flexible general purpose decoder for dcc
//            here: init, mainloop, globals
//
//
// ====== >>>> config.h  is the central definition file for the project
//
//
// content:   A DCC-Decoder for ATmega8515 and other AVR
//
//             1. Defines and variable definitions
//             2. Init
//             3. DCC Parser
//             5. MAIN: analyze command, call the action, do programming
//             6. Test and Simulation
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>		// put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "myeeprom.h"				// eeprom wrapper
#include "config.h"					// general definitions the decoder, cv's
#include "hardware.h"				// port definitions for target
#include "dcc_receiver.h"		// receiver for dcc
#include "dcc_decode.h"			// decode dcc and handle CV access
#include "dimm_engine.h"		// handling of output and feedback, LED control
#include "main.h"

// Kim beg
#if (SIMULATION != 0)
	#warning  SIMULATION activated

	#if (SIMULATION == 1)
		void simulat_receive(void);		// prototype
	#endif
#endif
// Kim end

//-----------------------------------------------------------------------------------
//
// Security Checks
//
#ifndef PROCESSOR
 #warning: there is no target processor defined - see hardware.h
 #warning: recommended: ATmega8 (8k Flash)
#endif

//-----------------------------------------------------------------------------------
//
// Global Data
//
//-----------------------------------------------------------------------------------

void init_main(void)
{
	DDRB	= (1<<SHIFT_SCK)		// out Schiebetakt 
				| (0<<SHIFT_MISO)		// inp Serieller Eingang
				| (1<<SHIFT_MOSI)		// out Serieller Ausgang
				| (1<<SHIFT_RCK)		// out Speichertakt ~SS
				| (1<<SHIFT_SCLR)		// out Reset fr Schieberegister  ~SCLR
				| (1<<SHIFT_OE);		// out Ausgangssteuerung 'OUTPUT ENABLE' ~OE

	DDRC  = 0xFF;							// All Bits as Output

	DDRD	=	 0								// inp
				|	 0								// inp
				| (0<<DCCIN)				// inp
				| (0<<JUMPER)				// inp
				| (0<<PROGTASTER)		// inp
				|	 0								// inp
				| (1<<LED)					// out, 1 turns on LED
				| (1<<DCC_ACK);			// out, sending 1 makes an ACK

	PORTB = (0<<SHIFT_SCK)		// out 0
				| (1<<SHIFT_MISO)		// inp pullup
				| (0<<SHIFT_MOSI)		// out 0
				| (0<<SHIFT_RCK)		// out 0	Den Speichertakt ~SS auf 0 halten
				| (1<<SHIFT_SCLR)		// out 1	SCLR am Schieberegister auf 1 stellen
				| (0<<SHIFT_OE);		// out 0

	PORTC = 0x00;							// out: all off     

	PORTD =	 1								// inp pullup
				|	 1								// inp pullup
				| (1<<DCCIN)				// inp pullup
				| (1<<JUMPER)				// inp pullup
				| (1<<PROGTASTER)		// inp pullup
				|	 1								// inp pullup
				| (0<<LED)					// out 0
				| (0<<DCC_ACK);			// out 0
}

//--------------------------------------------------------------------------------------------------
//
// here we check out whether we should save a current state of the decoder
//
// CurrentTarget should no be saved immediately after a command, so we do some timeout 

#define SAVE_TIMEOUT  (3000000L / TICK_PERIOD)
#if (SAVE_TIMEOUT == 0)
	#define SAVE_TIMEOUT  1
#endif

#if (SAVE_TIMEOUT > 240)
	#warning SAVE_TIMEOUT  too large
#endif

unsigned char	CheckSaveState;
signed char		CheckSave_timerval;

// saves:
unsigned char RAM_LastState[NUM_SIGNAL];

void CheckSave(void)
//============
{
	uint8_t sign;

	switch (CheckSaveState) {
		default:
		case 0:
			for (sign = 0; sign < NUM_SIGNAL; sign++) {
				RAM_LastState[sign] = my_eeprom_read_byte(&CV.LastPortState[sign]);
				CurrentBlink[sign]	= my_eeprom_read_byte(&CV.LastBlinkState[sign]);
			}
			CheckSave_timerval = timerval;
			CheckSaveState = 1;
			break;
		case 1:
			if (semaphor_get(C_DoSave)) {
				if (JUMPER_FITTED) {
					CheckSave_timerval = timerval;
					CheckSaveState = 2;
				}
			}
			break;
		case 2:
			if ((unsigned char)(timerval - CheckSave_timerval) > SAVE_TIMEOUT) {
				unsigned char my_mode;
				turn_led_on();
				my_mode = my_eeprom_read_byte(&CV.MODE);

				switch (my_mode) {
					default:
					case 32:
						for (sign = 0; sign < NUM_SIGNAL; sign++) {
							RAM_LastState[sign] = CurrentTarget[sign];
						}
					break;
				}

				for (sign = 0; sign < NUM_SIGNAL; sign++) {
					if (RAM_LastState[sign] != my_eeprom_read_byte(&CV.LastPortState[sign]))
						my_eeprom_write_byte(&CV.LastPortState[sign], RAM_LastState[sign]); 
					if (CurrentBlink[sign] != my_eeprom_read_byte(&CV.LastBlinkState[sign]))
						my_eeprom_write_byte(&CV.LastBlinkState[sign], CurrentBlink[sign]); 
				}

				if (Display_MainSign != my_eeprom_read_byte(&CV.DsplyMainSign)) {
					my_eeprom_write_byte(&CV.DsplyMainSign, Display_MainSign);
				} 
				if (Display_PrevSign != my_eeprom_read_byte(&CV.DsplyPrevSign)) {
					my_eeprom_write_byte(&CV.DsplyPrevSign, Display_PrevSign);
				} 

				CheckSaveState = 1;
				turn_led_off();
			} else {
				if (semaphor_get(C_DoSave)) {
					if (JUMPER_FITTED) {
						CheckSave_timerval = timerval;		// retrigger timer
				}
			}
		}
		break;
	}
}	// CheckSave

//------------------------------------------------------------------------
// This Routine is called when PROG is pressed
// -- manual programming and accordingly setting of CV's
//
#define DEBOUNCE  (50000L / TICK_PERIOD)
#if (DEBOUNCE == 0)
	#define DEBOUNCE  1
#endif

void DoProgramming(void)
//================
{
	unsigned char myCommand;
	signed char my_timerval;
	uint8_t	sign;
	my_timerval = timerval;

	while(timerval - my_timerval < DEBOUNCE);			// wait

	if (PROG_PRESSED) {														// still pressed?
		turn_led_on();

		while(PROG_PRESSED);												// wait for release

		my_timerval = timerval;
		while(timerval - my_timerval < DEBOUNCE);		// wait

		while(!PROG_PRESSED) {
			if (semaphor_get(C_Received)) {						// Message received
				if (analyze_message(&incoming)) {				// yes, any accessory
// Kim beg
					if ((my_eeprom_read_byte(&CV.Config) & 0x60) == 0x60) {
						// extended accessory
						my_eeprom_write_byte(&CV.myAddrH, (unsigned char)(ReceivedAddr >> 8) & 0x7F);
						my_eeprom_write_byte(&CV.myAddrL, (unsigned char)((ReceivedAddr >> 2) & 0x3F) + 1);
					} else {
						// basic accessory
						my_eeprom_write_byte(&CV.myAddrH, (unsigned char)(ReceivedAddr >> 6) & 0b00000111);
						my_eeprom_write_byte(&CV.myAddrL, (unsigned char) ReceivedAddr & 0b00111111);     
					}
//					myCommand = ReceivedCommand & 0x07;
					myCommand = ReceivedCommand & 0x1F;		// Aspect
					simple_preset(myCommand);

					if (JUMPER_FITTED) {
						for (sign = 0; sign < NUM_SIGNAL; sign++) {
							my_eeprom_write_byte(&CV.LastPortState[sign], 0xC0);	// Fahrbegriff Halt
						}
					}
// Kim end
					do {} while (!eeprom_is_ready());		// wait for write to complete
					LED_OFF;

					// we got reprogrammed ->
					// forget everthing running and restart decoder!
					cli();

					// laut diversen Internetseiten sollte folgender Code laufen -
					// tuts aber nicht, wenn man das Assemblerlistung ansieht.
					// void (*funcptr)( void ) = 0x0000;	// Set up function pointer
					// funcptr();													// Jump to Reset vector 0x0000

					__asm__ __volatile 
					(
						"ldi r30,0"	"\n\t"
						"ldi r31,0"	"\n\t"
						"icall"			"\n\t"
					);
					// return;
				} // if
			} // if
		}	// while

		turn_led_off();
		my_timerval = timerval;
		while(timerval - my_timerval < DEBOUNCE);		// wait    
		while(PROG_PRESSED);												// wait for release
	} // if
	return;
}	// DoProgramming

//--------------------------------------------------------------------------------------------

#define ASM_DUMMY() __asm__ __volatile__ ("" : : )

int main(void)
{
	init_main();					// setup hardware ports (to do!!)
	init_dimm_engine();		// setup timers and states
	init_dcc_receiver();	// setup dcc receiver
	init_dcc_decode();

	if (my_eeprom_read_byte(&CV.myAddrH) & 0x80)
		flash_led_fast(5);	// warning - we are unprogrammed
//	else
//		flash_led_fast(2);	// dimmer is running
	sei();								// Global enable interrupts
	CheckSaveState = 0;		// force reinit

// Kim beg
	#if (SIMULATION != 0)
		#if (SIMULATION == 3)
			action(8, 0);			// aspect, signNbr
			while(1) {
				if (semaphor_get(C_Tick)) {
					Dimmer();
				} 
			}
		#else
			simulat_receive();
		#endif
	#else
// Kim end
		unsigned char my_mode;

		while(1) {
			if (semaphor_query(C_Received)) {
				if (analyze_message(&incoming) >= 2) {		// MyAdr or greater received
					my_mode = my_eeprom_read_byte(&CV.MODE);
					switch (my_mode) {
						#if (PORT_ENABLED == TRUE)
						case 0:
							break;
						#endif

						case 32:
							action(ReceivedCommand, ReceivedActivate);
							break;
						default:
							break;
					}
				}
				semaphor_get(C_Received);									// now take away the protection
			}

			if (semaphor_get(C_Tick)) {
				Dimmer();
			} 

			CheckSave();

			if (PROG_PRESSED) {
				DoProgramming();
			}
		}
	#endif	// SIMULATION == 0
	return (0);
}

//===============================================================================
//
// Simulationen:
//
// 0: keine Simulation
// 1: Test der Empfangrootine; hierzu ein DCC Generator (aus opendcc/dccout.c)
// 2: Test der Portengine
//
#if (SIMULATION == 1)
//                                        Adressbyte   Datenbyte 
//                                         10AAAAAA    1aaaSCCR    // note: aaa = ~AAA
unsigned char message_adr001_out0_g[] = {0b10000001, 0b11111001};
unsigned char message_adr005_out1_g[] = {0b10000101, 0b11111011};
unsigned char message_adr380_out0_g[] = {0b10111100, 0b10101001};  // 380 = 0x17C = 0b101111100

// upstream interface:
struct
{
	unsigned char size;
	unsigned char dcc[5];
} next_message;

volatile unsigned char next_message_count;

enum do_states		// actual state
{
	dos_idle,
	dos_send_preamble,
	dos_send_bstart,
	dos_send_byte,
	dos_send_xor
};

struct
{
	enum do_states state;
	unsigned char ibyte;							// current index of byte in message
	unsigned char bits_in_state;			// Bits to output in this state
	unsigned char cur_byte;						// current byte
	unsigned char xor_byte;						// actual xor
	unsigned char current_dcc[5];			// current message in output processing
	unsigned char bytes_in_message;		// current size of message (decremented)
	unsigned char phase;
} doi;

unsigned char dccbit;

void do_send(unsigned char mydccbit)
{
	dccbit = mydccbit;
}

void dcc_bit_generator(void)
{
	switch (doi.state) {
		case dos_idle:
			do_send(1);
			if (next_message_count > 0) {
				memcpy(doi.current_dcc, next_message.dcc, sizeof(doi.current_dcc));
				doi.bytes_in_message = next_message.size;
				// no size checking - if (doi.cur_size > 5) doi.cur_size = 5;
				next_message_count--;
				doi.ibyte = 0;
				doi.xor_byte = 0;
				doi.bits_in_state = 14;
				doi.state = dos_send_preamble;
			}
			break;

		case dos_send_preamble:
			do_send(1);
			doi.bits_in_state--;
			if (doi.bits_in_state == 0)
				doi.state = dos_send_bstart;
			break;

		case dos_send_bstart:
			do_send(0);
			if (doi.bytes_in_message == 0) {		// message done, goto xor
				doi.cur_byte = doi.xor_byte;
				doi.state = dos_send_xor;
				doi.bits_in_state = 8;
			} else {														// get next addr or data
				doi.bytes_in_message--;
				doi.cur_byte = doi.current_dcc[doi.ibyte++];
				doi.xor_byte ^= doi.cur_byte;
				doi.state = dos_send_byte;
				doi.bits_in_state = 8;
			}
			break;

		case dos_send_byte:
			if (doi.cur_byte & 0x80)
				do_send(1);
			else
				do_send(0);
			doi.cur_byte <<= 1;
			doi.bits_in_state--;
			if (doi.bits_in_state == 0) {
				doi.state = dos_send_bstart;
			}
			break;

		case dos_send_xor:
			if (doi.cur_byte & 0x80)
				do_send(1);
			else
				do_send(0);
			doi.cur_byte <<= 1;
			doi.bits_in_state--;
			if (doi.bits_in_state == 0) {
				doi.state = dos_idle;
			}
			break;
	}
}

void dcc_generate_init(void)
{
	doi.state = dos_idle;
}

void simulat_receive(void)
{
	dcc_generate_init();
	dcc_bit_generator();
	dcc_bit_generator();

	// {preamble} 0 10AAAAAA 0 0AAA0AA1 0 000XXXXX 0 EEEEEEEE 1
	// BitStrm(ac2): P15 83 23 0F AF
	//unsigned char message_adr1293_asp15[] = {0x83, 0x23, 0x0F};
	//unsigned char message_adr1293_asp07[] = {0x83, 0x23, 0x07};
	//unsigned char message_adr1294_asp03[] = {0x83, 0x25, 0x03};
	unsigned char message_adr1292_asp04[] = {0x83, 0x21, 0x04};

	memcpy(next_message.dcc, message_adr1292_asp04, sizeof(doi.current_dcc));
	//next_message.size = 2;
	next_message.size	 = 3;
	next_message_count = 2;

	while(1) {
		dcc_bit_generator();
		dcc_receive();

		if (semaphor_query(C_Received)) {
			if (analyze_message(&incoming) >= 2) {		// MyAdr or greater received
				action(ReceivedCommand, ReceivedActivate);
			}
			semaphor_get(C_Received);
		}

		if (semaphor_query(C_DoSave)) {
			semaphor_get(C_DoSave);
		} 

		if (semaphor_get(C_Tick)) {
			Dimmer();
		} 

		//if (PROG_PRESSED) DoProgramming();         
	}	// while
}

#endif
//------------------------------------------------------------------------
